from pathspec import PathSpec
from pathspec.patterns import GitWildMatchPattern
from pathlib import Path
import toml


def read_text(path: Path) -> str:
    """Assuming all non-textual files are in the turbo_docs.toml ignore list"""
    try:
        with open(path, "r") as file:
            return file.read()
        
    except UnicodeDecodeError:
        with open(path, 'r', encoding='utf-8', errors='replace') as file:
            return file.read()
        
    except Exception as e:
        raise e


def collect_text_from_files(dir_path: Path, pathspec: PathSpec) -> dict:
    """recursively collect text from all files in a directory"""
    result = {}
    for path in dir_path.iterdir():
        if path.is_file():
            if not pathspec.match_file(str(path)):
                result[path] = read_text(path)
        elif path.is_dir():
            result.update(collect_text_from_files(path, pathspec))

    return result


def get_repo_text_dict():
    """Return a dictionary of all text in the current repo"""
    if not Path("turbo_docs.toml").exists():
        ignored_patterns = []
        print("Warning: 'turbo_docs.toml' not found. All files will be included.")
        print("See https://github.com/voynow/turbo-docs/blob/main/turbo_docs.toml for an example.")
    else:
        config = toml.load("turbo_docs.toml")
        ignored_patterns = config.get("ignore", [])

    pathspec = PathSpec.from_lines(GitWildMatchPattern, ignored_patterns)
    return collect_text_from_files(Path("."), pathspec)


def convert_dict_to_string(file_dict: dict):
    text = ""
    for path, content in file_dict.items():
        text += f"{path}:\n\n{content}\n\n"
    return text




TEMPLATE = """
"You are an expert software developement assistant. Write a README.md for the following repo:
{repo}

Here are the requirements for the README.md:
- written in highly formatted markdown
- badeges for github stars and pypi if applicable
- an utline written assuming the user knows nothing about the repo
- a section on why use this repo, pitching the repo to the user
- repo structure formatted as a tree
- example usage of important functions, especially functions exposed to the user, in code blocks
- emoji are encouraged. One for each section at minimum.
"""

def readme(repo, model, template=TEMPLATE):
    """
    Chose between GPT-3.5 Turbo and GPT-4, allow for template override, and
    generate a README.md file for the current repo.
    """
    from turbo_docs.utils import openai_api

    readme = "README.md"
    response, _ = openai_api.gpt_completion(template, model, repo=repo)
    
    with open(readme, "w", encoding="utf-8") as readme_file:
        readme_file.write(response)



import ast
from pathlib import Path
import shutil

TEMPLATE = """
You are an expert software developement assistant tasked with writing documentation for a repository.

The documentation must following the following format:
## function: function_name
#### args: arg1, arg2, ...
Insert description of the function here
<end>

In your description, describe the function in english as concisely as possible in at least one sentence but less than four sentences. Subtly describe the usefule/counterintuitive aspects of the function. Do not list arguments. Do not give code examples. Minimize tokens at all costs.

Here is the function:
{function}

Now, write the documentation following the format above.
"""

def generate_docs(path, function, model="gpt-4", template=TEMPLATE, chain=None):
    """
    Generate and write documentation for a function. Use existing chain for
    a cohesive chain of gpt completions.
    """
    from turbo_docs.utils import openai_api

    documentation, chain = openai_api.gpt_completion(
        model=model, chain=chain, template=template, function=function
    )
    doc_path = Path("docs") / path.with_suffix(".md")
    doc_path.parent.mkdir(parents=True, exist_ok=True)
    with open(doc_path, "a") as f:
        f.write(f"{documentation}\n\n")
    return chain


def docs(repo_dict, model, template=TEMPLATE):
    """Parse code for functions and generate documentation for each"""
    chain = None

    # remove old docs since we are appending
    if Path("docs").exists():
        shutil.rmtree("docs")

    for path, content in repo_dict.items():
        if path.suffix == ".py":
            module = ast.parse(content)
            functions = [
                (node, ast.unparse(node))
                for node in module.body
                if isinstance(node, ast.FunctionDef)
            ]
            for node, function in functions:
                print(f"Generating docs for {path.stem}.{node.name}")
                chain = generate_docs(
                    path, function, model=model, template=template, chain=chain
                )



import click
from pathlib import Path
import pyperclip
import tiktoken
from turbo_docs.commands import readme as readme_module
from turbo_docs.commands import docs as docs_module
from turbo_docs.utils import directory, cli_options

def resolve_model(gpt3):
    if gpt3:
        model = "gpt-3.5-turbo-16k"
    else:
        model = "gpt-4"
        print(
            "Warning: This model is under limited beta access and is not available to all users."
        )
        print("Warning: GPT-4 api calls tend to be slower than other models.")

    print(f"Using model: {model}")
    return model

def num_tokens_from_string(string: str, encoding_name: str = "cl100k_base") -> int:
    """Returns the number of tokens in a text string."""
    encoding = tiktoken.get_encoding(encoding_name)
    num_tokens = len(encoding.encode(string, allowed_special="all"))
    return num_tokens

@click.command()
@cli_options.copy
@cli_options.readme
@cli_options.gpt3
@cli_options.docs
def driver(
    copy: bool,
    readme: bool,
    gpt3: bool,
    docs: bool,
) -> None:
    """
    Pull text from all files in the current directory and apply the following commands:

    'turbo_docs --copy' copies the text to clipboard
        Useful for wokring with ChatGPT

    'turbo_docs --readme' generates a README.md file
        Useful for keeping documentation up to date

    'turbo_docs --gpt3' uses GPT-3.5-turbo-16k
        Useful if you don't have GPT-4 access

    'turbo_docs --docs' generates a docs for each file
        Useful for keeping documentation up to date
    """
    dir_text_dict = directory.get_repo_text_dict()

    if readme:
        if Path("README.md") in dir_text_dict:
            del dir_text_dict[Path("README.md")]
        dir_text_str = directory.convert_dict_to_string(dir_text_dict)
        model = resolve_model(gpt3)
        readme_module.readme(dir_text_str, model)
        print("Generated README.md")

    if docs:
        model = resolve_model(gpt3)
        docs_module.docs(dir_text_dict, model)
        print("Generated documentation")

    if copy:
        dir_text_str = directory.convert_dict_to_string(dir_text_dict)
        num_tokens = num_tokens_from_string(dir_text_str)
        pyperclip.copy(dir_text_str)
        print(f"Directory text copied to clipboard containing {num_tokens} tokens")



import click
from typing import Callable

def copy(func: Callable) -> Callable:
    return click.option(
        '--copy',
        default=False,
        is_flag=True,
        help='Copy the directory text to clipboard'
    )(func)

def readme(func: Callable) -> Callable:
    return click.option(
        '--readme',
        default=False,
        is_flag=True,
        help='Generate README.md file'
    )(func)

def gpt3(func: Callable) -> Callable:
    return click.option(
        '--gpt3',
        default=False,
        is_flag=True,
        help='Use the GPT-3.5 Turbo model'
    )(func)

def docs(func: Callable) -> Callable:
    return click.option(
        '--docs',
        default=False,
        is_flag=True,
        help='Generate documentation for all code files'
    )(func)



import os
import openai

if not os.environ.get("OPENAI_API_KEY"):
	print("OpenAI API key is not set. Please set it as an environment variable (export OPENAI_API_KEY=<your_api_key>) or enter it below.")
	print("If you have not done so already, create an OpenAI account at https://platform.openai.com/overview.")
	os.environ['OPENAI_API_KEY'] = input("Secret key: ")

from llm_blocks import chat_utils


def gpt_completion(template, model="gpt-4", chain=None, **inputs):
	"""Generic chat wrapper over the OpenAI API"""

	if chain is None:
		chain = chat_utils.GenericChain(
			template=template, 
			model_name=model
		)
	try:
		resp = chain(inputs)['text']
	except openai.error.InvalidRequestError as e:
		raise ValueError(f"Caught OpenAI API error: {e}")
	return resp, chain



from setuptools import setup, find_packages

setup(
	name="turbo_docs",
	version="1.0.4",
	packages=find_packages(),
	install_requires=[
		"openai",
		"click",
		"pyperclip",
    	"toml",
        "pathspec",
        "llm-blocks",
        "tiktoken"
	],
	entry_points={
		"console_scripts": [
			"turbo_docs=turbo_docs.generate:driver"
		],
	},
	classifiers=[
		"Development Status :: 3 - Alpha",
		"Intended Audience :: Developers",
		"License :: OSI Approved :: MIT License",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: 3.9",
	],
	long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type='text/markdown'
)

# python setup.py sdist bdist_wheel
# twine upload dist/*